
/* main_flex.c
   (C) 2020 Yann Guidon whygee@f-cpu.org
   The source code of this project is released under the AGPLv3 or later,
   see https://www.gnu.org/licenses/agpl-3.0.en.html

This file is an example "main loop" for a server that can sit idle
and decide to switch to a polling method if it receives external input.

Compile:
$ gcc -DHTTAP_VERBOSE -DSHOW_CLIENT -DHTTAP_SERVE_FILES -DENABLE_LOOPBACK -Wall main_flex.c -o main_flex

run:
$ ./main_flex
*/

// globals to control the "work"
long int cnt=0;
char firstchar='a';
char maxchar='z';

#define HTTAP_TIMEOUT_S (10)  // s
#define HTTAP_FAST_MS   (50)  // ms
#define HTTAP_SLOW_MS   (500) // ms

// Use the provided signal() code
#define HTTAP_USE_SIGNAL
#include "HTTaP.h"

#include <poll.h>

struct pollfd my_pollfd={
  0, // stdin
  POLLIN, // only for reading
  0
};

/*
typedef enum {
// When there is host work to do, don't wait at the poll() call (timeout=0)
  H_POLL_SLOW = 1, // Poll slowly when there is no interactive activity
  H_POLL_FAST = 2, // Poll more often when some activity has been detected

// When there is no work to do, "block" at the poll() call
  H_BLOCKING  = 4, // Activity has been recently received, timeout has not yet elapsed.
    // => timeout is recalculated based on previous timestamp gotten from clock_gettime().
  H_BLOCKED   = 8  // No connection or timeout, hard block and wait for signal
     // or incoming order.  => timeout = -1.
} PollState_e;
PollState_e PollState=H_POLL_SLOW;
*/

struct itimerval itv_poll;
struct timespec timestamp;

int H_poll_timeout; // millisecond waited in poll()

void change_state_POLL_SLOW () {
  PollState = H_POLL_SLOW;
  H_poll_timeout=0;
  // about 500ms latency to wake up :
  itv_poll.it_value.tv_usec = HTTAP_SLOW_MS*1000;
  VERBOSE_PRINTF("H_POLL_SLOW\n");
}

void change_state_POLL_FAST () {
  PollState = H_POLL_FAST;
  H_poll_timeout=0;
  // about 50ms between polls if active :
  itv_poll.it_value.tv_usec = HTTAP_FAST_MS*1000;
  VERBOSE_PRINTF("H_POLL_FAST\n");
}

void change_state_BLOCKING () {
  timeout_counter=1;  // only loop once
  PollState = H_BLOCKING;
  VERBOSE_PRINTF("H_BLOCKING\n");
}

void change_state_BLOCKED () {
  H_poll_timeout=-1;  // just wait.
  timeout_counter=1;  // only loop once
  PollState = H_BLOCKED;
  VERBOSE_PRINTF("H_BLOCKED\n");
}

// User-visible functions to change the
// mode without interfering with the timeout:

void HTTaP_mode_block() {
  if (PollState == H_POLL_FAST)
    change_state_BLOCKING();
  else
    if (PollState == H_POLL_SLOW)
      change_state_BLOCKED();
}

void HTTaP_mode_poll() {
  if (PollState == H_BLOCKED)
    change_state_POLL_SLOW();
  else
    if (PollState == H_BLOCKING)
      change_state_POLL_FAST();
}

int HTTaP_server2(int timeout) {
  int l, ret=0;  // return 1 if useful activity has happened.
  char buff[1<<5];

  int p=poll( &my_pollfd, 1, timeout?0:H_poll_timeout);
  if (p==-1) {
    if (errno != EINTR)
      show_exit("\npoll() error ", 1);
  }
  if (p < 1) {
//    printf(" poll() : nothing new... \n");
    if (timeout) {
      printf(" close ! \n");
    }
    return 0;
  }

  if (my_pollfd.revents & ( POLLPRI | POLLHUP | POLLNVAL))
    show_exit("\npoll() .revents flags: WTF ?\n", 0);

  if (my_pollfd.revents & POLLIN ){
    l=read(0, &buff[0], sizeof(buff));
    if (l==-1)
      show_exit("\nread() from stdin", 1);
    printf("\nread %d bytes\n", l);

    if (l==2) {
      switch(buff[0]) {
        case 'q': // Quit
          abort_program = 1;
          timeout_counter=1;
          break;

        case 's': // Slow
          HTTaP_mode_poll();
          break;

        case 'f': // Fast
          ret=1;
          break;

        case 'p': // Pause
          ret=1;
          HTTaP_mode_block();
          break;

        case '0': //
          HTTaP_mode_poll();
          ret=1;
          cnt=0;
          break;

        case '1': //
          ret=1;
          firstchar='a';
          break;

        case '2': //
          ret=1;
          firstchar='r';
          break;

        case '3': //
          ret=1;
          maxchar='z';
          break;

        case '4': //
          ret=1;
          maxchar='t';
          break;

        case '9': //
          cnt=10000000;
          ret=1;
          break;
      }
    }

    // flush the buffer
    while (l == sizeof(buff)) {
      l=read(0, &buff[0], sizeof(buff));
      if (l==-1)
        show_exit("\nread() from stdin", 1);
    }
  }

  return ret;
}

/////////////////////// High-level wrappers ///////////////////////

void init_HTTaP() {
  __sighandler_t e=signal(SIGALRM, timer_handler);
  if ((e!=SIG_IGN) && (e!=SIG_DFL))
    show_exit("\nsignal(): realtime signal is already busy", 1);
  intercept_signals();

  bzero(&itv_poll, sizeof(itv_poll));
  clock_gettime(ITIMER_REAL, &timestamp);
  change_state_POLL_SLOW();
}

// a wrapper that manages the timeouts :
void poll_HTTaP() {
  struct timespec current_time;
  int timediff, timeout=0;

  timeout_counter=0; // let the work be done by default.

  if (PollState & (H_BLOCKING | H_POLL_FAST) ) {
    // get the (rough) time elapsed since the last successful command:
    clock_gettime(ITIMER_REAL, &current_time);
    timediff = timestamp.tv_sec - current_time.tv_sec;

    // check if the server must close the socket if no input:
    if (timediff <= 0) {
      timeout = 1;
      if (PollState == H_BLOCKING)
        change_state_BLOCKED();
      else {
//      if (PollState == H_POLL_FAST) : implicit
          change_state_POLL_SLOW();
      }
    }
    else {
      if (PollState == H_BLOCKING) {
        H_poll_timeout = timediff * 1000;
//      printf(" waiting %d ms  \n", H_poll_timeout);
      }
    }
  }

  // run the server once:
  if (HTTaP_server2(timeout)) {
    // valid command received: restart the timeout counters !
    clock_gettime(ITIMER_REAL, &timestamp);
    timestamp.tv_sec += HTTAP_TIMEOUT_S;

    if (PollState == H_POLL_SLOW)
      change_state_POLL_FAST();
    else
      if (PollState == H_BLOCKED)
        change_state_BLOCKING();
  }

  if (PollState & (H_POLL_FAST | H_POLL_SLOW)) {
    if (setitimer(ITIMER_REAL, &itv_poll, NULL) == -1)
      show_exit("\n restarting a timer", 1);
  }
  else
    timeout_counter=1; // just let the host check its own stuff
                       // before halting during the next poll()
}

/////////////////////// User code ///////////////////////

// just eat up CPU time
void some_work() {
  printf("   %ld  ", cnt++);

  if ( cnt > 200000 )
    HTTaP_mode_block();
  else {
    char c=firstchar;
    while (c <= maxchar) {
      putchar(c++);
      fflush(NULL);
    }
    putchar(0xD);
    fflush(NULL);
  }
}

int main(int argc, char *argv[]) {
  init_HTTaP();

  do {
    poll_HTTaP(); // must precede the user's "work"

    do
      some_work();
    while (!timeout_counter);
  } while (!abort_program);

  printf("bye from main() !");
}
